#include <Arduino.h>
#include <SoftwareSerial.h>
#include <MD_MAX72xx.h>
#include <EEPROM.h>
#include <RTClib.h>
#include "fonts.h"

#define VERSION F("v. 1.0.0")

#define SERIAL_SPEED 9600
#define SERIAL_BT_SPEED 9600
#define SERIAL_BT_STATE_PIN 3
#define SERIAL_BT_TX_PIN 4
#define SERIAL_BT_RX_PIN 5
#define LED_HW_TYPE MD_MAX72XX::FC16_HW
#define LED_CS_PIN 6
#define LED_MODULES 4

#define STATE_TIME 0
#define STATE_DATE 1
#define STATE_TEMP 2

// config addresses
#define ADDR_CHECKSUM 0x0
#define ADDR_INTENSITY 0x1
#define ADDR_TIME_ENABLED_INTERVAL 0x2
#define ADDR_DATE_ENABLED_INTERVAL 0x3
#define ADDR_TEMP_ENABLED_INTERVAL 0x4
#define ADDR_DATE_ENABLED 0x5
#define ADDR_TEMP_ENABLED 0x6
#define ADDR_SCROLL_SPEED 0x7
#define ADDR_SCROLL_WAIT 0x8

// config default values
#define CFG_CHECKSUM 0x2
#define CFG_INTENSITY_DEFAULT 0xf
#define CFG_TIME_ENABLED_INTERVAL_DEFAULT 35
#define CFG_DATE_ENABLED_INTERVAL_DEFAULT 3
#define CFG_TEMP_ENABLED_INTERVAL_DEFAULT 3
#define CFG_DATE_ENABLED_DEFAULT 1
#define CFG_TEMP_ENABLED_DEFAULT 1
#define CFG_SCROLL_SPEED_DEFAULT 5
#define CFG_SCROLL_WAIT_DEFAULT 40

// config shortcut macros
#define CFG_INTENSITY EEPROM.read(ADDR_INTENSITY)
#define CFG_TIME_ENABLED_INTERVAL (((uint16_t)EEPROM.read(ADDR_TIME_ENABLED_INTERVAL)) * 1000)
#define CFG_DATE_ENABLED_INTERVAL (((uint16_t)EEPROM.read(ADDR_DATE_ENABLED_INTERVAL)) * 1000)
#define CFG_TEMP_ENABLED_INTERVAL (((uint16_t)EEPROM.read(ADDR_TEMP_ENABLED_INTERVAL)) * 1000)
#define CFG_DATE_ENABLED EEPROM.read(ADDR_DATE_ENABLED)
#define CFG_TEMP_ENABLED EEPROM.read(ADDR_TEMP_ENABLED)
#define CFG_SCROLL_SPEED (((uint16_t)EEPROM.read(ADDR_SCROLL_SPEED)) * 10)
#define CFG_SCROLL_WAIT (((uint16_t)EEPROM.read(ADDR_SCROLL_WAIT)) * 100)

// global variables
SoftwareSerial btSerial(SERIAL_BT_RX_PIN, SERIAL_BT_TX_PIN);
MD_MAX72XX led(LED_HW_TYPE, LED_CS_PIN, LED_MODULES);
RTC_DS3231 rtc;

#define BUFFER_SIZE 20
char buffer[BUFFER_SIZE];

uint8_t state = STATE_TIME;
unsigned long lastClockCall = 0;
unsigned long lastScrollCall = 0;

// functions
uint8_t conv2d(const char *p, uint8_t maxLen)
{
  uint8_t v = 0;
  for (uint8_t i = 0; *(p + i) != '\0' && maxLen-- > 0; i++)
  {
    v = (v * 10) + (*(p + i) - '0');
  }
  return v;
}

void setup()
{
  // setup pins
  pinMode(SERIAL_BT_RX_PIN, INPUT);
  pinMode(SERIAL_BT_TX_PIN, OUTPUT);
  pinMode(LED_CS_PIN, OUTPUT);

  // setup serial
  Serial.begin(SERIAL_SPEED);

  // setup bluetooth serial
  btSerial.begin(SERIAL_BT_SPEED);

  // setup RTC
  rtc.begin();

  // setup led matrix
  led.begin();

  // clear flash
  if (EEPROM.read(ADDR_CHECKSUM) != CFG_CHECKSUM)
  {
    EEPROM.write(ADDR_CHECKSUM, CFG_CHECKSUM);
    EEPROM.write(ADDR_INTENSITY, CFG_INTENSITY_DEFAULT);
    EEPROM.write(ADDR_TIME_ENABLED_INTERVAL, CFG_TIME_ENABLED_INTERVAL_DEFAULT);
    EEPROM.write(ADDR_DATE_ENABLED_INTERVAL, CFG_DATE_ENABLED_INTERVAL_DEFAULT);
    EEPROM.write(ADDR_TEMP_ENABLED_INTERVAL, CFG_TEMP_ENABLED_INTERVAL_DEFAULT);
    EEPROM.write(ADDR_DATE_ENABLED, CFG_DATE_ENABLED_DEFAULT);
    EEPROM.write(ADDR_TEMP_ENABLED, CFG_TEMP_ENABLED_DEFAULT);
    EEPROM.write(ADDR_SCROLL_SPEED, CFG_SCROLL_SPEED_DEFAULT);
    EEPROM.write(ADDR_SCROLL_WAIT, CFG_SCROLL_WAIT_DEFAULT);
  }

  // clear led
  led.clear();
  led.control(MD_MAX72XX::INTENSITY, CFG_INTENSITY);
  led.update(MD_MAX72XX::OFF); // no automatic updates wanted

  lastClockCall = lastScrollCall = millis();
}

// command processing
#define COMMAND_SIZE 255
char command[COMMAND_SIZE];
bool clearCommand = true;

char *readCommand()
{
  // if clearCommand flag was set, clear the command buffer before we receive new data
  if (clearCommand)
  {
    clearCommand = false;
    memset(command, 0, sizeof(command) / sizeof(command[0]));
  }

  // current length of the command
  uint8_t commandLength = strnlen(command, COMMAND_SIZE);

  // receive available bytes
  while (btSerial.available() > 0)
  {
    // check for command overflow
    if (commandLength == COMMAND_SIZE)
    {
      clearCommand = true; // clear the command buffer and pointers next time
      return NULL;         // no command is available for processing
    }

    command[commandLength] = btSerial.read();                             // receive one byte
    if (command[commandLength] == '\r' || command[commandLength] == '\n') // the received byte is end-of-command identifier
    {
      // end of command was received, null terminate the string, the \r or \n is not needed as part of the command
      command[commandLength] = '\0';
      if (strnlen(command, COMMAND_SIZE) == 0) // if the command length is 0, we did not receive anything else but the \r or \n, ignore
      {
        clearCommand = true;
        return NULL;
      }

      // first 3 bytes are the command (e.g. SET, ADJ, ), then 0 terminate, this is just for convenience
      // possible arguments to the command will be from command[4] onwards. Maybe all of command[4] and onwards will be NULL if the command
      // had no parameters. This needs to be checked in code consuming the command
      command[3] = '\0';

      // next time we need to start with a fresh command buffer
      clearCommand = true;

      // return that we received a command
      return command;
    }

    commandLength++; // the command buffer is now one character longer
  }

  // no complete command was received yet
  return NULL;
}

uint8_t charBuf[8];
#define SCROLL_BUFF_SIZE 255
char scrollBuf[SCROLL_BUFF_SIZE];
bool showText(char *text, MD_MAX72XX::fontType_t *font, uint8_t tsd, uint16_t scrollSkip = 0)
{
  led.setFont(font);
  uint16_t position = led.getColumnCount() - 1;
  bool center = scrollSkip == 0;

  // render the text
  while (*text != '\0')
  {
    uint8_t width = led.getChar((uint8_t)*text, 8, charBuf);
    for (uint8_t i = 0; i < width; i++)
    {
      if (scrollSkip > 0)
      {
        // if we are still about to skip columns, reduce the scrollSkip and do nothing else
        scrollSkip--;
      }
      else
      {
        // we are not skipping characters to scroll, so now we can actually draw
        led.setColumn(position, charBuf[i]);
        if (position == 0)
        {
          break; // there is no more space on the led matrix
        }
        else
        {
          position--; // move to the next column
        }
      }
    }

    if (position == 0)
    {
      break; // there is no more space on the led matrix
    }

    text++; // go to next char

    if (scrollSkip > 0)
    {
      scrollSkip--;
    }
    else
    {
      position--; // move to the next column (one column space between characters)
    }
  }

  if (scrollSkip > 0)
  {
    // we are done scrolling, but we had not enough text to draw anything (text completely scrolled out of the screen)
    return true;
  }

  // center the text if it is shorter than the display (centering is only enabled when not scrolling)
  while (position > 0 && center)
  {
    led.transform(MD_MAX72XX::TSR);
    position = min(position - 1, position - 2); // avoid underflow of position
  }

  // translate downwards if requested
  while (tsd-- > 0)
  {
    led.transform(MD_MAX72XX::TSD);
  }

  return false;
}

void showClock()
{
  if (rtc.lostPower())
  {
    buffer[0] = buffer[1] = buffer[2] = buffer[3] = '-';
    buffer[4] = '\0';
    showText(buffer, NULL, 0);
    lastClockCall = millis();
    return;
  }
  unsigned long diff = millis() - lastClockCall;
  DateTime now = rtc.now();

  if (state == STATE_TIME) // display the time
  {
    snprintf_P(buffer, BUFFER_SIZE, PSTR("%02d:%02d:%02d"), now.hour(), now.minute(), now.second());
    showText(buffer, font_fw_3x5, 1);

    if (diff >= CFG_TIME_ENABLED_INTERVAL)
    {
      lastClockCall = millis();
      if (CFG_DATE_ENABLED)
      {
        state = STATE_DATE;
      }
      else if (CFG_TEMP_ENABLED)
      {
        state = STATE_TEMP;
      }
      else
      {
        state = STATE_TIME;
      }
    }
  }
  else if (state == STATE_DATE) // display the date
  {
    snprintf_P(buffer, BUFFER_SIZE, PSTR("%02d.%02d."), now.day(), now.month());
    showText(buffer, font_fw_3x5, 1);

    if (diff >= CFG_DATE_ENABLED_INTERVAL)
    {
      lastClockCall = millis();
      if (CFG_TEMP_ENABLED)
      {
        state = STATE_TEMP;
      }
      else
      {
        state = STATE_TIME;
      }
    }
  }
  else if (state == STATE_TEMP) // display the temperature
  {
    snprintf_P(buffer, BUFFER_SIZE, PSTR("%02d °C"), (int)rtc.getTemperature());
    showText(buffer, font_fw_3x5, 1);

    if (diff >= CFG_TEMP_ENABLED_INTERVAL)
    {
      lastClockCall = millis();
      state = STATE_TIME;
    }
  }
}

bool scrollEnabled = false;
uint16_t scrollSkip;
uint16_t scrollWait;

void beginScrollText(const char *txt)
{

  strncpy(scrollBuf, txt, SCROLL_BUFF_SIZE);
  scrollEnabled = true;
  scrollSkip = 0;
  scrollWait = CFG_SCROLL_WAIT;
  lastScrollCall = millis();
  showText(scrollBuf, NULL, 0);
}

void endScrollText()
{
  scrollEnabled = false;
  memset(scrollBuf, 0, sizeof(scrollBuf) / sizeof(scrollBuf[0]));
}

void scrollText()
{
  if (!scrollEnabled)
  {
    return;
  }

  unsigned long diff = millis() - lastScrollCall;
  if (scrollWait > 0)
  {
    lastScrollCall = millis();
    scrollWait = scrollWait - min(scrollWait, diff);
    if (scrollWait > 0)
    {
      showText(scrollBuf, NULL, 0, scrollSkip);
      return;
    }
  }

  if (diff >= CFG_SCROLL_SPEED)
  {
    bool scrollDone = showText(scrollBuf, NULL, 0, scrollSkip++);
    if (scrollDone)
    {
      scrollWait = CFG_SCROLL_WAIT;
      scrollSkip = 0;
      showText(scrollBuf, NULL, 0);
    }
    lastScrollCall = millis();
  }
  else
  {
    showText(scrollBuf, NULL, 0, scrollSkip);
  }
}

bool blink = false;
bool blinkState = false;
unsigned long lastBlinkCall;

void startBlink()
{
  if (blink == false)
  {
    blink = true;
    lastBlinkCall = millis();
  }
}

void stopBlink()
{
  blink = false;
}

void doBlink()
{
  if (!blink)
  {
    return;
  }

  if (blinkState)
  {
    led.transform(MD_MAX72XX::TINV);
  }

  unsigned long diff = millis() - lastBlinkCall;
  if (diff > 1000)
  {
    blinkState = !blinkState;
    lastBlinkCall = millis();
  }
}

DateTime *nextTime = 0;

void loop()
{
  // clear the output
  led.clear();

  if (!digitalRead(SERIAL_BT_STATE_PIN)) // BT is not connected
  {
    showClock();

    // indicate that there is no bluetooth connection
    led.setPoint(7, 0, true);
  }
  else // BT is connected
  {
    // read command
    char *cmd = readCommand();
    if (cmd != NULL) // command received?
    {
      if (strncmp_P(cmd, PSTR("MTC"), 3) == 0)
      {
        btSerial.print(F("MTC "));
        btSerial.println(VERSION);
      }
      else if (strncmp_P(cmd, PSTR("ADJ"), 3) == 0)
      {
        // Adjust the rtc clock, arguments from position 4 "2000-01-01T00:00:00"
        cmd[24] = '\0';
        rtc.adjust(DateTime(&cmd[4]));
        btSerial.println(F("ADJ"));
      }
      else if (strncmp_P(cmd, PSTR("SET"), 3) == 0) // SET a config value
      {
        // Set a setting value, arguments:
        // - form position 4 to 5 contains the setting name (6 is the \0 byte)
        // - from position 7 the new value begins
        cmd[6] = '\0';
        if (strncmp_P(&cmd[4], PSTR("IN"), 2) == 0) // intensity
        {
          uint8_t intensity = conv2d(&cmd[7], 2);
          intensity = (intensity == 0) ? intensity = 1 : ((intensity > MAX_INTENSITY) ? MAX_INTENSITY : intensity);
          EEPROM.update(ADDR_INTENSITY, intensity);
          led.control(MD_MAX72XX::INTENSITY, intensity);
          btSerial.println(F("SET IN"));
        }
        else if (strncmp_P(&cmd[4], PSTR("TI"), 2) == 0) // time interval
        {
          uint8_t interval = conv2d(&cmd[7], 3);
          EEPROM.update(ADDR_TIME_ENABLED_INTERVAL, interval);
          btSerial.println(F("SET TI"));
        }
        else if (strncmp_P(&cmd[4], PSTR("DI"), 2) == 0) // date interval
        {
          uint8_t interval = conv2d(&cmd[7], 3);
          EEPROM.update(ADDR_DATE_ENABLED_INTERVAL, interval);
          btSerial.println(F("SET TI"));
        }
        else if (strncmp_P(&cmd[4], PSTR("CI"), 2) == 0) // temperature interval
        {
          uint8_t interval = conv2d(&cmd[7], 3);
          EEPROM.update(ADDR_TEMP_ENABLED_INTERVAL, interval);
          btSerial.println(F("SET CI"));
        }
        else if (strncmp_P(&cmd[4], PSTR("DE"), 2) == 0) // date enabled
        {
          EEPROM.update(ADDR_DATE_ENABLED, cmd[7] == '1');
          btSerial.println(F("SET DE"));
        }
        else if (strncmp_P(&cmd[4], PSTR("CE"), 2) == 0) // temperature enabled
        {
          EEPROM.update(ADDR_TEMP_ENABLED, cmd[7] == '1');
          btSerial.println(F("SET CE"));
        }
        else if (strncmp_P(&cmd[4], PSTR("SS"), 2) == 0) // scroll speed
        {
          uint8_t speed = conv2d(&cmd[7], 3);
          EEPROM.update(ADDR_SCROLL_SPEED, speed);
          btSerial.println(F("SET SS"));
        }
        else if (strncmp_P(&cmd[4], PSTR("SW"), 2) == 0) // scroll wait
        {
          uint8_t interval = conv2d(&cmd[7], 3);
          EEPROM.update(ADDR_SCROLL_WAIT, interval);
          btSerial.println(F("SET SW"));
        }
        else
        {
          btSerial.println(F("ERR"));
        }
      }
      else if (strncmp_P(cmd, PSTR("GET"), 3) == 0) // GET a config value
      {
        // Get a setting value, arguments:
        // - form position 4 to 5 contains the setting name (6 is the \0 byte)
        cmd[6] = '\0';
        if (strncmp_P(&cmd[4], PSTR("IN"), 2) == 0) // intensity
        {
          btSerial.print(F("GET IN "));
          btSerial.println(CFG_INTENSITY);
        }
        else if (strncmp_P(&cmd[4], PSTR("TI"), 2) == 0) // time interval
        {
          btSerial.print(F("GET TI "));
          btSerial.println(CFG_TIME_ENABLED_INTERVAL / 1000);
        }
        else if (strncmp_P(&cmd[4], PSTR("DI"), 2) == 0) // date interval
        {
          btSerial.print(F("GET DI "));
          btSerial.println(CFG_DATE_ENABLED_INTERVAL / 1000);
        }
        else if (strncmp_P(&cmd[4], PSTR("CI"), 2) == 0) // temperature interval
        {
          btSerial.print(F("GET CI "));
          btSerial.println(CFG_TEMP_ENABLED_INTERVAL / 1000);
        }
        else if (strncmp_P(&cmd[4], PSTR("DE"), 2) == 0) // date enabled
        {
          btSerial.print(F("GET DE "));
          btSerial.println(CFG_DATE_ENABLED);
        }
        else if (strncmp_P(&cmd[4], PSTR("CE"), 2) == 0) // temperature enabled
        {
          btSerial.print(F("GET CE "));
          btSerial.println(CFG_TEMP_ENABLED);
        }
        else if (strncmp_P(&cmd[4], PSTR("SS"), 2) == 0) // scroll speed
        {
          btSerial.print(F("GET SS "));
          btSerial.println(CFG_SCROLL_SPEED / 10);
        }
        else if (strncmp_P(&cmd[4], PSTR("SW"), 2) == 0) // scroll wait
        {
          btSerial.print(F("GET SW "));
          btSerial.println(CFG_TEMP_ENABLED_INTERVAL / 100);
        }
        else
        {
          btSerial.println(F("ERR"));
        }
      }
      else if (strncmp_P(cmd, PSTR("SCR"), 3) == 0) // Scroll Text
      {
        // position 4 is set to 0 and from position 5 the text to scroll starts
        // it might be empty when we want to stop the scroll
        if (cmd[4] == '\0')
        {
          endScrollText(); // do not scroll text anymore
        }
        else
        {
          // scroll text should start with additional space
          cmd[3] = ' ';
          beginScrollText(&cmd[3]);
        }
        btSerial.println(F("SCR"));
      }
      else if (strncmp_P(cmd, PSTR("NXT"), 3) == 0) // Next Time
      {
        // position 4 is set to 0 and from position 5 the time will start
        // Next appointmen starts at this time. Format "2000-01-01T00:00:00"
        if (nextTime != 0)
        {
          delete nextTime;
        }
        cmd[24] = '\0';
        nextTime = new DateTime(&cmd[4]);

        btSerial.println(F("NXT"));
      }
      else // everything else is unknown (ERR)
      {
        btSerial.println(F("ERR"));
      }
    }

    if (strnlen(scrollBuf, SCROLL_BUFF_SIZE) > 0)
    {
      scrollText();
    }
    else
    {
      showClock();
    }

    if (nextTime != 0)
    {
      TimeSpan ts = *nextTime - rtc.now();
      uint8_t devices = led.getDeviceCount();
      if (ts.totalseconds() < -30)
      {
        // time is up
        // 30 seconds past the time, stop blinking and delete time
        delete nextTime;
        nextTime = 0;
        stopBlink();
        // also now we can stop scrolling
        memset(scrollBuf, 0, sizeof(scrollBuf) / sizeof(scrollBuf[0]));
      }
      else if (ts.totalseconds() < 30)
      {
        // only 30 seconds left, alarm by blinking
        startBlink();
      }
      else if (ts.totalseconds() < ((devices + 1) * 60))
      {
        // only some minutes left, start warning
        for (uint16_t device = 0; device < devices; device++)
        {
          if (ts.totalseconds() < (device + 1) * 60)
          {
            led.transform(device, device, MD_MAX72XX::TINV);
          }
        }
      }
    }
  }

  doBlink();
  led.update();
  delay(5); // sleep some tme
}
